/*	$NetBSD$	*/

/*
 * Copyright (c) 2001 ITOH Yasufumi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <signal.h>	/* sigset_t */

#define DCMEM_RETRYCNT	10

#define DCMEM_BLK_SZ		512
#define DCMEM_SIZE		256

#define DCMEM_START_PROGRAM	0	/* XXX fixed? */
#define DCMEM_SIZE_PROGRAM	128	/* XXX fixed? */

#define DCMEM_FAT_LOC	254
#define DCMEM_DEFAULT_ROOTBLK_LOC	255

/*
 * on-flash root block
 */
struct dcmem_rootblk {
	u_int8_t	dr_mark[16];
#define DCMEM_RB_MARK	\
	"\125\125\125\125\125\125\125\125\125\125\125\125\125\125\125\125"

	u_int8_t	dr_custom_color;
#define DCMEM_USE_STANDARD_COLOR	0
#define DCMEM_USE_CUSTOM_COLOR		1
	u_int8_t	dr_color_blue, dr_color_green, dr_color_red,
			dr_color_alpha;
	u_int8_t	dr_unused1[27];

	u_int8_t	dr_timestamp[8];	/* BCD timestamp */
	u_int8_t	dr_unused2[8];

	u_int16_t	dr_maxblk;	/* max block # (little endian) */
	u_int16_t	dr_minblk;	/* min block # (little endian) */
	u_int16_t	dr_rbloc;	/* rootblock location (little endian) */
	u_int16_t	dr_fatloc;	/* FAT location (little endian) */
	u_int16_t	dr_fatsz;	/* FAT size (little endian) */
	u_int16_t	dr_dirloc;	/* directory location (little endian) */
	u_int16_t	dr_dirsz;	/* directory size (little endian) */
	u_int16_t	dr_icon;	/* icon shape (little endian) */
	u_int16_t	dr_nblk;	/* # user blocks (little endian) */

	u_int8_t	dr_unknown1[6];	/* 1F 00 00 00 80 00 */

	u_int8_t	dr_unused3[424];
};

/*
 * on-flash directory entry
 */
struct dcdirent {
	u_int8_t	dd_type;	/* file type */
#define DCMEM_FTYPE_NOFILE	0
#define DCMEM_FTYPE_DATA	0x33
#define DCMEM_FTYPE_PROGRAM	0xCC

	u_int8_t	dd_copyprot;	/* copy protection */
#define DCMEM_COPY_OK		0
#define DCMEM_COPY_PROTECTED	0xFF

	u_int16_t	dd_start;	/* start block # (little endian) */
	u_int8_t	dd_name[12];	/* file name */
	u_int8_t	dd_timestamp[8];/* BCD timestamp */
	u_int16_t	dd_size;	/* size in blocks (little endian) */
	u_int16_t	dd_hdroff;	/* off hdr in blocks (little endian) */
	u_int8_t	dd_unused[4];
};

/*
 * on-memory structure
 */
struct dcmem {
	int		dm_fd;		/* file descriptor */

	int		dm_flags;
#define DCMEM_FRDONLY		1	/* read only */

	int		dm_writing;
	sigset_t	dm_osigset;

	u_int16_t	dm_maxblk;	/* max block # */
	u_int16_t	dm_minblk;	/* min block # */
	u_int16_t	dm_rbloc;	/* rootblock location */
	u_int16_t	dm_fatloc;	/* FAT location */
	u_int16_t	dm_fatsz;	/* FAT size */
	u_int16_t	dm_dirloc;	/* directory location */
	u_int16_t	dm_dirsz;	/* directory size */
	u_int16_t	dm_nblk;	/* # user blocks */

	u_int16_t	*dm_fat;
	struct dcdirent	*dm_dir;

	int		dm_rblk_changed;
	struct dcmem_rootblk dm_rblk;
};

#define DCMEM_FAT_END		0xFFFA
#define DCMEM_FAT_FREE		0xFFFC
#define DCMEM_FAT_BROKEN	0xFFFF	/* ??? I don't know correct value */

#define DCMEM_FAT_IS_FREE(m, fatno)	\
	((m)->dm_fat[fatno] == htole16(DCMEM_FAT_FREE))

#define DCMEM_FAT_MARK(m, fatno, n)	\
	((m)->dm_fat[fatno] = htole16(n))

/*
 * VMI file
 */
struct dcvmi {
	u_int8_t	di_chkbyte[4];
#define DCMEM_VMI_CHKAND0	0123
#define DCMEM_VMI_CHKAND1	0105
#define DCMEM_VMI_CHKAND2	0107
#define DCMEM_VMI_CHKAND3	0101

	u_int8_t	di_descr[32];
	u_int8_t	di_copyright[32];

	u_int16_t	di_year;
	u_int8_t	di_month;
	u_int8_t	di_day;
	u_int8_t	di_hour;
	u_int8_t	di_minute;
	u_int8_t	di_second;
	u_int8_t	di_weekday;	/* 0: Sunday, ..., 6: Saturday */

	u_int16_t	di_ver;
#if 0	/* ??? I see 0 and 0x0100 */
#define DCMEM_VMI_VERSION	0
#endif
	u_int16_t	di_filenum;
#define DCMEM_VMI_FILENUM	1

	u_int8_t	di_resource[8];	/* .VMS resource name */
#define DCMEM_VMS_SUFFIX	".VMS"

	u_int8_t	di_fn[12];	/* on-flash filename */

	u_int16_t	di_mode;
#define DCMEM_VMI_MODE_PROGRAM		2
#define DCMEM_VMI_MODE_COPY_PROTECTED	1

	u_int16_t	di_pad;		/* ? */
	u_int32_t	di_size;	/* size in byte */
};

/*
 * library error codes
 */
#define DCMEM_ESYS		-1	/* (see errno) */
#define DCMEM_EFORMAT		1	/* not formatted */
#define DCMEM_ENOTSUPP		2	/* un-supported format */
#define DCMEM_ERBLK		3	/* root block corrupt */
#define DCMEM_EINCONSISTENT	4	/* structure inconsistent */
#define DCMEM_ERDONLY		5	/* write requeted on read-only mode */
#define DCMEM_EUXEOF		6	/* unexpected end-of-file */
#define DCMEM_EHWPROBLEM	7	/* memory card hardware problem */
#define DCMEM_ENOENT		8	/* file not found */
#define DCMEM_EEXISTS		9	/* file already exists */
#define DCMEM_ENOSPC		10	/* no space */
#define DCMEM_EDIRFULL		11	/* no free directory slot */
#define DCMEM_EINVAL		12	/* invalid argument */

int dcmem_fat_find_free_data __P((const struct dcmem *));
int dcmem_fat_cnt_free_data __P((const struct dcmem *));
int dcmem_fat_cnt_free_program __P((const struct dcmem *));
int dcmem_fat_find_free_sys __P((const struct dcmem *));
int dcmem_fat_cnt_free_sys __P((const struct dcmem *));
int dcmem_fat_next __P((const struct dcmem *, int fatno));
void dcmem_fat_free_chain_data __P((const struct dcmem *, int fatno));

struct dcdirent *dcmem_dir_enumerate __P((const struct dcmem *, int *ppos));
struct dcdirent *dcmem_dir_lookup_name __P((const struct dcmem *m, const u_int8_t *name));
int dcmem_dir_remove_name __P((struct dcmem *, const u_int8_t *name));
struct dcdirent *dcmem_dir_alloc __P((struct dcmem *, const u_int8_t *name, int *perr));

int dcmem_read_file_dirent __P((const struct dcmem *, const struct dcdirent *, void *));
int dcmem_read_file_name __P((const struct dcmem *, const u_int8_t *name, void *buf, struct dcdirent *pd));

int dcmem_write_file_dirent __P((struct dcmem *, const struct dcdirent *, void *));

int dcmem_read_1 __P((const struct dcmem *, void *ptr, int blk));
struct dcmem *dcmem_open __P((const char *fn, int flags, int *perr));
int dcmem_read_chain __P((const struct dcmem *, void *buf,
			int top, int sz, int is_sys));
int dcmem_close __P((struct dcmem *));

void dcmem_start_write __P((struct dcmem *));
void dcmem_end_write __P((struct dcmem *));
int dcmem_write_1 __P((struct dcmem *, const void *ptr, int blk));
int dcmem_write_data __P((struct dcmem *, const void *ptr, int sz, int *pfatstart));
int dcmem_write_program __P((struct dcmem *, const void *ptr, int sz));
int dcmem_update_sys __P((struct dcmem *, void *ptr, int isfat));
int dcmem_emergency_save __P((struct dcmem *));
#define DCMEM_SAVE_TEMPLATE	"/var/tmp/dcmemgXXXXXXXX"
int dcmem_flush __P((struct dcmem *));

const char * dcmem_strerror __P((int));
